@@ -45,6 +45,7 @@
 #include "eeprom_i2c.h"
 #include "eeprom_spi.h"
 #include "flash_cfi.h"
+#include "yx5200.h"
 #include "megasd.h"
 
 /* Cart database entry */
@@ -71,6 +72,7 @@ static void mapper_flash_w8(uint32 address, uint32 data);
 static void mapper_flash_w16(uint32 address, uint32 data);
 static uint32 mapper_flash_r8(uint32 address);
 static uint32 mapper_flash_r16(uint32 address);
+static void mapper_kaiserwave_w(uint32 address, uint32 data);
 static uint32 mapper_smw_64_r(uint32 address);
 static void mapper_smw_64_w(uint32 address, uint32 data);
 static void mapper_realtec_w(uint32 address, uint32 data);
@@ -105,7 +107,7 @@ static const md_entry_t rom_database[] =
 {
 /* Funny World & Balloon Boy */
   {0x0000,0x06ab,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},1,0,NULL,m68k_unused_8_w,NULL,mapper_realtec_w}},
-/* Whac-a-Critter */
+/* Whac-a-Critter - Mallet Legend */
   {0xffff,0xf863,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},1,0,NULL,m68k_unused_8_w,NULL,mapper_realtec_w}},
 /* Earth Defense */
   {0xffff,0x44fb,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},1,0,NULL,m68k_unused_8_w,NULL,mapper_realtec_w}},
@@ -133,11 +135,9 @@ static const md_entry_t rom_database[] =
 /* RADICA (Sensible Soccer Plus edition) (PAL) */
   {0x0000,0x1f7f,0x00,0x00,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},0,0,mapper_128k_radica_r,m68k_unused_8_w,NULL,NULL}},
 
-
 /* Tenchi wo Kurau III: Sangokushi Gaiden - Chinese Fighter */
   {0x9490,0x8180,0x40,0x6f,{{0x00,0x00,0x00,0x00},{0xf0000c,0xf0000c,0xf0000c,0xf0000c},{0x400000,0x400004,0x400008,0x40000c},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_alt_regs_w}},
 
-
 /* Top Fighter */
   {0x4eb9,0x5d8b,0x60,0x7f,{{0x00,0x00,0x00,0x00},{0xf00007,0xf00007,0xf00007,0xffffff},{0x600001,0x600003,0x600005,0x000000},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_regs_w}},
 /* Soul Edge VS Samurai Spirits  */
@@ -150,110 +150,109 @@ static const md_entry_t rom_database[] =
   {0x0000,0x507c,0x60,0x7f,{{0x00,0x00,0x00,0x00},{0xf00007,0xf00007,0xf00007,0xffffff},{0x600001,0x600003,0x600005,0x000000},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_regs_w}},
 /* Super King Kong 99 */
   {0x0000,0x7d6e,0x60,0x7f,{{0x00,0x00,0x00,0x00},{0xf00007,0xf00007,0xf00007,0xffffff},{0x600001,0x600003,0x600005,0x000000},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_regs_w}},
+/* Super Donkey Kong 99 (bootleg version of Super King Kong 99 ? all protection checks seem patched except one, resulting in broken end game sequence if hardware is not implemented) */
+  {0x0000,0x16dc,0x60,0x7f,{{0x00,0x00,0x00,0x00},{0xf00007,0xf00007,0xf00007,0xffffff},{0x600001,0x600003,0x600005,0x000000},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_regs_w}},
 /* Gunfight 3-in-1 */
   {0x0000,0x6ff8,0x60,0x7f,{{0x00,0x00,0x00,0x00},{0xf00007,0xf00007,0xf00007,0xffffff},{0x600001,0x600003,0x600005,0x000000},0,1,NULL,m68k_unused_8_w,default_regs_r,custom_regs_w}}, 
 /* Pokemon Stadium */
   {0x0000,0x843c,0x70,0x7f,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},0,1,NULL,m68k_unused_8_w,NULL,custom_regs_w}},
 
-
 /* Tekken 3 Special (original dump) (a bootleg version also exists, with patched protection & different boot routine which reads unused !TIME mapped area) */
   {0x0000,0xc2f0,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,tekken_regs_r,tekken_regs_w}},
 
-
 /* Lion King 2 */
   {0xffff,0x1d9b,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
 /* Squirell King */
   {0x0000,0x8ec8,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
 /* Tiny Toon Adventures 3 */
   {0x2020,0xed9c,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
-/* Lian Huan Pao - Barver Battle Saga (registers accessed by Z80, related to sound engine ?) */
+/* Barver Battle Saga - Tai Kong Zhan Shi (registers accessed by Z80, related to sound engine ?) */
   {0x30b9,0x1c2a,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
 /* Shui Hu Zhuan (registers accessed by Z80, related to sound engine ?) */
   {0x6001,0x0211,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
 /* Feng Shen Ying Jie Chuan (registers accessed by Z80, related to sound engine ?) */
   {0xffff,0x5d98,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
-/* (*) Shui Hu - Feng Yun Zhuan (patched ROM, unused registers) */
+/* (*) Shui Hu: Feng Yun Zhuan (patched ROM, unused registers) */
   {0x3332,0x872b,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xfffffd,0xfffffd,0xffffff,0xffffff},{0x400000,0x400004,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,default_regs_w}},
 
-
 /* (*) Chao Ji Da Fu Weng (patched ROM, various words witten to register, long word also read from $7E0000, unknown banking hardware ?) */
   {0xa697,0xa697,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,NULL,default_regs_w}},
 
-/* (*) Aq Renkan Awa (patched ROM, ON/OFF bit sequence is written to register, unknown banking hardware ?) */
+/* (*) Aq Renkan Awa - A Q Lian Huan Pao (patched ROM, ON/OFF bit sequence is written to register, unknown banking hardware ?) */
   {0x8104,0x0517,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400001,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,NULL,default_regs_w}},
 
-
 /* (*) Tun Shi Tian Di III (patched ROM, unused register) */
   {0x0000,0x9c5e,0x40,0x40,{{0xab,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400046,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 
-
-/* Ma Jiang Qing Ren - Ji Ma Jiang Zhi */
+/* Chao Ji Mahjong Club (uses 16-bits reads)  */
+  {0x0000,0x2d5c,0x40,0x40,{{0x90,0xd3,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r_16,NULL}},
+/* Ma Jiang Qing Ren: Ji Ma Jiang Zhi */
   {0x0000,0x7037,0x40,0x40,{{0x90,0xd3,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x401000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
-/* Super Majon Club */
+/* Du Shen Zhi Meng Huan Poker */
   {0x0000,0x3b95,0x40,0x40,{{0x90,0xd3,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x401000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Feng Kuang Tao Hua Yuan (original version from Creaton Softec Inc) (a bootleg version also exists with patched protection and minor title screen variations) */
   {0x0000,0x9dc4,0x40,0x40,{{0x90,0xd3,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x401000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
+/* (*) Zhan Qi - Chinese Battle Chess (patched ROM ? registers values are read and compared to expected values but comparison results are unused) */
+  {0x0000,0x4704,0x40,0x40,{{0x90,0xd3,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x401000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 
-
-/* (*) Jiu Ji Ma Jiang II - Ye Yan Bian (patched ROM, using expected register value - $0f - crashes the game) (uses 16-bits reads) */
+/* (*) Jiu Ji Ma Jiang II: Ye Yan Bian (patched ROM, using expected register value $f0 crashes the game) (uses 16-bits reads) */
   {0x0c44,0xba81,0x40,0x40,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r_16,NULL}},
-/* 16 Zhang Ma Jiang (uses 16-bits reads) */
+/* Shi Liu Zhang Ma Jiang - 16 Tiles Mahjong (uses 16-bits reads) */
   {0xfb40,0x4bed,0x40,0x40,{{0x00,0xaa,0x00,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x400002,0x000000,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r_16,NULL}},
-/* 16 Tiles Mahjong II (uses 16-bits reads) */
+/* Shi Liu Zhang Ma Jiang II - 16 Tiles Mahjong II (uses 16-bits reads) */
   {0xffff,0x0903,0x40,0x40,{{0x00,0x00,0xc9,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x400004,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r_16,NULL}},
-/* Thunderbolt II (uses 16-bits reads) */
+/* Lei Dian Chuan Shuo II - Thunderbolt II (uses 16-bits reads) */
   {0x0000,0x1585,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r_16,NULL}},
 
-
-/* Chaoji Puke - Super Poker (correct ROM dump, original release is an overdump) */
+/* Chao Ji Poker (correct ROM dump, original release is an overdump) */
   {0xffff,0xd7b0,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Super Bubble Bobble */
   {0x0000,0x16cd,0x40,0x40,{{0x55,0x0f,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
-/* Tenchi wo Kurau II - The Battle of Red Cliffs (Unl) */
+/* San Guo Yan Yi: Huo Shao Chi Bi - The Battle of Red Cliffs */
   {0x0000,0xed61,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Huan Le Tao Qi Shu - Smart Mouse */
   {0x0000,0x1a28,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
+/* Sichuan Style Mahjong */
+  {0x0000,0x71e1,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* (*) Hei Tao 2 - Super Big 2 (patched ROM, unused registers) */
   {0x0000,0x5843,0x40,0x40,{{0x55,0x0f,0xaa,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Mighty Morphin Power Rangers - The Fighting Edition */
   {0x0000,0x2288,0x40,0x40,{{0x55,0x0f,0xc9,0x18},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
-/* Elf Wor */
+/* Ling Huan Dao Shi - Super Magican */
   {0x0080,0x3dba,0x40,0x40,{{0x55,0x0f,0xc9,0x18},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Ya-Se Chuanshuo */
   {0xffff,0xd472,0x40,0x40,{{0x63,0x98,0xc9,0x18},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
-/* 777 Casino (For first one, 0x55 works as well. Other values are never used so they are guessed from on other unlicensed games using similar mapper) */
+/* Meng Huan Shui Guo Pan - 777 Casino (For first one, 0x55 works as well. Other values seem never used so they are guessed from other unlicensed games using similar mapper) */
   {0x0000,0xf8d9,0x40,0x40,{{0x63,0x98,0xc9,0x18},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Wu Kong Wai Zhuan (original) (a bootleg version also exists, with patched protection & modified SRAM test routine ?) */
   {0x0000,0x19ff,0x40,0x40,{{0x63,0x98,0xc9,0x18},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Soul Blade */
   {0x0000,0x0c5b,0x40,0x40,{{0x63,0x98,0xc9,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
-/* Shi San Zhang Ma Jiang: Zhong Guo Mei Nv Pian (only $400004 is checked, other addresses/values are assumed) */
+/* Shi San Zhang Ma Jiang: Zhong Guo Mei Nv Pian (only $400004 seem checked, for which 0xaa works as well, other addresses/values are guessed from other unlicensed games using similar mapper) */
   {0xffff,0xd41d,0x40,0x40,{{0x63,0x98,0xc9,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
+/* Rockman X3 (original version) (only two last addresses seem checked, for which 0xaa/0x18 work as well, other addresses/values are guessed from other unlicensed games using similar mapper) */
+  {0xffff,0x6fdd,0x40,0x40,{{0x63,0x98,0xc9,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0x400000,0x400002,0x400004,0x400006},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 
 /* King of Fighter 98 */
   {0x0000,0xd0a0,0x48,0x4f,{{0x00,0x00,0xaa,0xf0},{0xffffff,0xffffff,0xfc0000,0xfc0000},{0x000000,0x000000,0x480000,0x4c0000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 
-
 /* Rock Heaven */
   {0x6cca,0x2395,0x50,0x50,{{0x50,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x500008,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 /* Rock World */
   {0x3547,0xa3da,0x50,0x50,{{0x50,0xa0,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x500008,0x500208,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,default_regs_r,NULL}},
 
-
 /* Rockman X3 (bootleg version ? two last register returned values are ignored, note that 0xaa/0x18 would work as well) */
   {0x0000,0x9d0e,0x40,0x40,{{0x0c,0x00,0xc9,0xf0},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0x000000,0x400004,0x400006},0,0,default_regs_r,m68k_unused_8_w,default_regs_r,NULL}},
 
-
 /* (*) Dragon Ball Final Bout  (patched ROM, in original code, different switches occurs depending on returned value $00-$0f) */
   {0xc65a,0xc65a,0x00,0x00,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0x000000,0x000000,0x000000},0,0,default_regs_r,m68k_unused_8_w,NULL,NULL}},
-/* (*) Yang Jia Jiang - Yang Warrior Family (patched ROM, register value unused) */
+/* (*) Yi Men Ying Lie: Yang Jia Jiang - Yang Warrior Family (patched ROM, register value unused) */
   {0x0000,0x96b0,0x00,0x00,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0x000000,0x000000,0x000000},0,0,default_regs_r,m68k_unused_8_w,NULL,NULL}},
 /* Super Mario 2 1998 */
   {0xffff,0x0474,0x00,0x00,{{0x0a,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0x000000,0x000000,0x000000},0,0,default_regs_r,m68k_unused_8_w,NULL,NULL}},
 /* Super Mario World */
   {0x2020,0xb4eb,0x00,0x00,{{0x1c,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0x000000,0x000000,0x000000},0,0,default_regs_r,m68k_unused_8_w,NULL,NULL}},
 
-
 /* King of Fighter 99 */
   {0x0000,0x021e,0x00,0x00,{{0x00,0x01,0x1f,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0xa13002,0xa1303e,0x000000},0,0,custom_regs_r,default_regs_w,NULL,NULL}},
 /* Pocket Monster */
@@ -263,11 +262,9 @@ static const md_entry_t rom_database[] =
 /* A Bug's Life (bootleg version ? two last register returned values are ignored & first register test has been modified ?) */
   {0x7f7f,0x2aad,0x00,0x00,{{0x28,0x01,0x1f,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0xa13000,0xa13002,0xa1303e,0x000000},0,0,default_regs_r,m68k_unused_8_w,NULL,NULL}},
 
-
 /* Game no Kanzume Otokuyou */
   {0x0000,0xf9d1,0x00,0x00,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},0,0,NULL,mapper_seganet_w,NULL,NULL}},
 
-
 /* Top Shooter (arcade hardware) */
   {0xffff,0x3632,0x20,0x20,{{0x00,0x00,0x00,0x00},{0xffffff,0xffffff,0xffffff,0xffffff},{0x000000,0x000000,0x000000,0x000000},0,0,NULL,m68k_unused_8_w,topshooter_r,topshooter_w}}
 };
@@ -539,9 +536,10 @@ void md_cart_init(void)
 
   /* detect specific mappers */
   else if (strstr(rominfo.international,"DEMONS OF ASTEBORG") ||
-           strstr(rominfo.international,"ASTEBROS"))
+           strstr(rominfo.international,"ASTEBROS") ||
+           strstr(rominfo.domestic,"DAEMON CLAW"))
   {
-    /* default SSF2+SRAM mapper (fixes Steam versions which incorrectly use Everdrive extended SSF mapper indicator in ROM header) */ 
+    /* default SSF2+SRAM mapper (fixes patched versions incorrectly indicating Everdrive extended SSF mapper in ROM header) */ 
     cart.hw.time_w = default_time_w;
   }
   else if (strstr(rominfo.consoletype,"SEGA SSF"))
@@ -661,24 +659,56 @@ void md_cart_init(void)
       zbank_memory_map[i].write  = zbank_unused_w;
     }
   }
-  else if (strstr(rominfo.ROMType,"GM") && strstr(rominfo.product,"00000000-42"))
+  else if (strstr(rominfo.ROMType,"GM") && strstr(rominfo.product,"00000000-42")) /* Escape 2042 */
   {
-    /* Flashkit-MD mapper (ROM only) */
+    /* initialize CFI flash memory hardware */
+    flash_cfi_init(M29W320EB, NULL);
+
+    /* Flashkit-MD mapper (4MB ROM only) */
     m68k.memory_map[0x00].read8   = mapper_flash_r8;
     m68k.memory_map[0x00].read16  = mapper_flash_r16;
     m68k.memory_map[0x00].write8  = mapper_flash_w8;
     m68k.memory_map[0x00].write16 = mapper_flash_w16;
     zbank_memory_map[0x00].read   = mapper_flash_r8;
     zbank_memory_map[0x00].write  = mapper_flash_w8;
-
-    /* initialize CFI flash memory device */
-    flash_cfi_init(M29W320EB);
   }
   else if (strstr(rominfo.ROMType,"GM") && strstr(rominfo.product,"00000000-00") && 
-           (((rominfo.checksum == 0xcdf5) && (rominfo.realchecksum = 0x603A)) ||  /* Life on Mars */
-            ((rominfo.checksum == 0x6bd5) && (rominfo.realchecksum = 0x1FEA))))   /* Life on Earth Reimagined */
+           (((rominfo.checksum == 0xcdf5) && (rominfo.realchecksum == 0x603a)) || /* Life on Mars */
+            ((rominfo.checksum == 0x6bd5) && (rominfo.realchecksum == 0x1fea))))  /* Life on Earth Reimagined */
   {
-    /* SGDK flash-save mapper (ROM+SRAM) */
+    /* initialize CFI flash memory hardware */
+    flash_cfi_init(S29GL064N_04, NULL);
+
+    /* custom flash-save mapper (4MB ROM with backup RAM mapped in last 64KB bank) */
+    /* Note: upper 4MB of flash memory is apparently not accessible */
+    m68k.memory_map[0x3f].base    = sram.sram;
+    m68k.memory_map[0x00].read8   = m68k.memory_map[0x3f].read8   = mapper_flash_r8;
+    m68k.memory_map[0x00].read16  = m68k.memory_map[0x3f].read16  = mapper_flash_r16;
+    m68k.memory_map[0x00].write8  = m68k.memory_map[0x3f].write8  = mapper_flash_w8;
+    m68k.memory_map[0x00].write16 = m68k.memory_map[0x3f].write16 = mapper_flash_w16;
+    zbank_memory_map[0x00].read   = zbank_memory_map[0x3f].read   = mapper_flash_r8;
+    zbank_memory_map[0x00].write  = zbank_memory_map[0x3f].write  = mapper_flash_w8;
+  }
+  else if (strstr(rominfo.ROMType,"GM") && strstr(rominfo.product,"00000000-00") && 
+           (((rominfo.checksum == 0x45c1) && (rominfo.realchecksum == 0xc613))))  /*  The Secret Of The Four Winds */
+  {
+    /* KAISER WAVE board !TIME handler */ 
+    cart.hw.time_w = mapper_kaiserwave_w;
+
+    /* enable YX5200 hardware */
+    cart.special |= HW_YX5200;
+
+    /* allocate & clear blip buffer for YX5200 audio stream */
+    snd.blips[3] = blip_new(snd.sample_rate / 10);
+    if (snd.blips[3]) blip_clear(snd.blips[3]);
+
+    /* initialize YX5200 audio */
+    yx5200_init(snd.sample_rate);
+
+    /* initialize CFI flash memory hardware */
+    flash_cfi_init(S29GL064N_04, NULL);
+
+    /* custom flash-save mapper (4MB ROM with backup RAM mapped in last 64KB bank) */
     m68k.memory_map[0x3f].base    = sram.sram;
     m68k.memory_map[0x00].read8   = m68k.memory_map[0x3f].read8   = mapper_flash_r8;
     m68k.memory_map[0x00].read16  = m68k.memory_map[0x3f].read16  = mapper_flash_r16;
@@ -687,8 +717,54 @@ void md_cart_init(void)
     zbank_memory_map[0x00].read   = zbank_memory_map[0x3f].read   = mapper_flash_r8;
     zbank_memory_map[0x00].write  = zbank_memory_map[0x3f].write  = mapper_flash_w8;
 
-    /* initialize CFI flash memory device */
-    flash_cfi_init(S29GL064N_04);
+    /* 64M extension mapping (upper 4MB of flash memory is apparently unused but still accessible as cartridge board uses same signals as described in https://gitlab.com/doragasu/flatmap64) */
+    for (i=0x40; i<0x80; i++)
+    {
+      m68k.memory_map[i].read8    = NULL;
+      m68k.memory_map[i].read16   = NULL;
+      m68k.memory_map[i].write8   = m68k_unused_8_w;
+      m68k.memory_map[i].write16  = m68k_unused_16_w;
+      zbank_memory_map[i].read    = zbank_unused_r;
+      zbank_memory_map[i].write   = zbank_unused_w;
+    }
+  }
+  else if (strstr(rominfo.international, "COLOCODX")) /* ColocoDX */
+  {
+    /* CFI flash memory OTP area (credit to Stargazer) */
+    const uint16 flash_otp[FLASH_OTP_AREA_SIZE] =
+    {
+      0x2923,0xBE84,0xE16C,0xD6AE,0x5290,0x49F1,0xF1BB,0xE9EB,0xB3A6,0xDB3C,0x870C,0x3E99,0x245E,0x0D1C,0x06B7,0x47DE,
+      0xB312,0x4DC8,0x43BB,0x8BA6,0x1F03,0x5A7D,0x0938,0x251F,0x5DD4,0xCBFC,0x96F5,0x453B,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,
+      0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF
+    };
+
+    /* initialize CFI flash memory hardware */
+    flash_cfi_init(S29GL064N_04, flash_otp);
+
+    /* SGDK flash-save mapper (8MB ROM with backup RAM mapped in one of the two last 64KB banks, see https://gitlab.com/doragasu/sgdk-flash-save) */
+    m68k.memory_map[0x7e].base    = sram.sram; /* SRAM initially mapped to bank #127 */
+    m68k.memory_map[0x00].read8   = m68k.memory_map[0x7f].read8   = m68k.memory_map[0x7e].read8   = mapper_flash_r8;
+    m68k.memory_map[0x00].read16  = m68k.memory_map[0x7f].read16  = m68k.memory_map[0x7e].read16  = mapper_flash_r16;
+    m68k.memory_map[0x00].write8  = m68k.memory_map[0x7f].write8  = m68k.memory_map[0x7e].write8  = mapper_flash_w8;
+    m68k.memory_map[0x00].write16 = m68k.memory_map[0x7f].write16 = m68k.memory_map[0x7e].write16 = mapper_flash_w16;
+    zbank_memory_map[0x00].read   = zbank_memory_map[0x7f].read   = zbank_memory_map[0x7e].read   = mapper_flash_r8;
+    zbank_memory_map[0x00].write  = zbank_memory_map[0x7f].write  = zbank_memory_map[0x7e].write  = mapper_flash_w8;
+
+    /* 64M extension mapping (see https://gitlab.com/doragasu/flatmap64) */
+    for (i=0x40; i<0x7e; i++)
+    {
+      m68k.memory_map[i].read8    = NULL;
+      m68k.memory_map[i].read16   = NULL;
+      m68k.memory_map[i].write8   = m68k_unused_8_w;
+      m68k.memory_map[i].write16  = m68k_unused_16_w;
+      zbank_memory_map[i].read    = zbank_unused_r;
+      zbank_memory_map[i].write   = zbank_unused_w;
+    }
   }
   else if ((cart.romsize == 0x400000) && 
            (READ_BYTE(cart.rom, 0x200150) == 'C') &&
@@ -844,9 +920,11 @@ void md_cart_init(void)
   /* force Mega CD sound hardware initialization when MegaSD overlay is enabled (if not already initialized)  */
   if ((cart.special & HW_MEGASD) && !snd.blips[1] && !snd.blips[2])
   {
-    /* allocate blip buffers for PCM and CD-DA audio streams */
+    /* allocate & clear blip buffers for PCM and CD-DA audio streams */
     snd.blips[1] = blip_new(snd.sample_rate / 10);
     snd.blips[2] = blip_new(snd.sample_rate / 10);
+    if (snd.blips[1]) blip_clear(snd.blips[1]);
+    if (snd.blips[2]) blip_clear(snd.blips[2]);
 
     /* initialize PCM and CD-DA audio */
     audio_set_rate(snd.sample_rate, snd.frame_rate);
@@ -888,6 +966,12 @@ void md_cart_reset(int hard_reset)
     megasd_reset();
   }
 
+  /* YX5200 hardware */
+  if (cart.special & HW_YX5200)
+  {
+    yx5200_reset();
+  }
+
   /* SVP chip */
   if (svp)
   {
@@ -981,6 +1065,12 @@ int md_cart_context_save(uint8 *state)
     bufferptr += flash_cfi_context_save(&state[bufferptr]);
   }
 
+  /* YX5200 hardware */
+  if (cart.special & HW_YX5200)
+  {
+    bufferptr += yx5200_context_save(&state[bufferptr]);
+  }
+
   return bufferptr;
 }
 
@@ -1049,6 +1139,12 @@ int md_cart_context_load(uint8 *state)
     bufferptr += flash_cfi_context_load(&state[bufferptr]);
   }
 
+  /* YX5200 hardware */
+  if (cart.special & HW_YX5200)
+  {
+    bufferptr += yx5200_context_load(&state[bufferptr]);
+  }
+
   return bufferptr;
 }
 
@@ -1542,6 +1638,23 @@ static uint32 mapper_flash_r16(uint32 address)
   return flash_cfi_read(address >> 1);;
 }
 
+/* 
+  KAISER WAVE board mapper
+*/
+static void mapper_kaiserwave_w(uint32 address, uint32 data)
+{
+  /* /UWR only */
+  if (!(address & 1))
+  {
+    /* YX5200 RX input line is mapped to D0 */
+    yx5200_write(data & 0x01);
+  }
+  else
+  {
+    m68k_unused_8_w(address, data);
+  }
+}
+ 
 /* 
   Super Mario World 64 (unlicensed) mapper
 */
