@@ -548,6 +548,9 @@ int vdp_context_load(uint8 *state)
   }
   else
   {
+    /* first restore VDP register #1 bit 2 value to prevent spurious Mode4->Mode5 VRAM switching */
+    reg[1] = (reg[1] & ~0x04) | (temp_reg[1] & 0x04);
+
     for (i=0;i<0x20;i++) 
     {
       vdp_reg_w(i, temp_reg[i], 0);
@@ -590,6 +593,30 @@ int vdp_context_load(uint8 *state)
     {
       color_update_m5(i, *(uint16 *)&cram[i << 1]);
     }
+
+    /* reinitialize rendering functions */
+    update_bg_pattern_cache = update_bg_pattern_cache_m5;
+    parse_satb = parse_satb_m5;
+    render_bg = (reg[11] & 0x04) ? (config.enhanced_vscroll ? render_bg_m5_vs_enhanced : render_bg_m5_vs) : render_bg_m5;
+    render_obj = (reg[12] & 0x08) ? render_obj_m5_ste : render_obj_m5;
+
+    /* reinitialize bus access functions */
+    vdp_68k_data_w = vdp_68k_data_w_m5;
+    vdp_z80_data_w = vdp_z80_data_w_m5;
+    vdp_68k_data_r = vdp_68k_data_r_m5;
+    vdp_z80_data_r = vdp_z80_data_r_m5;
+
+    /* clear HVC latched value (unless HVC latch bit is set) */
+    if (!(reg[0] & 0x02))
+    {
+      hvc_latch = 0;
+    }
+
+    /* restore vertical counter max value */
+    vc_max = vc_table[(reg[1] >> 2) & 3][vdp_pal];
+
+    /* indicate display height change should be applied on next frame */
+    bitmap.viewport.changed |= 2; 
   }
   else
   {
@@ -1771,19 +1798,35 @@ static void vdp_reg_w(unsigned int r, unsigned int d, unsigned int cycles)
         /* Mega Drive VDP only */
         if (system_hw & SYSTEM_MD)
         {
-          int i;
+          int i, j;
+          uint8 temp[0x400];
+
           if (d & 0x04)
           {
+            /* 16K->64K address decoding */
+            for (i=0; i<0x4000; i+=0x400)
+            {
+              /* make temporary copy of 1KB VRAM chunk */
+              memcpy(temp, vram + i, 0x400);
+
+              /* re-arrange VRAM addressing (first and last words are unchanged) */
+              for (j=2; j<0x3fe; j+=2)
+              {
+                *(uint16 *)(vram + i + ((j << 1) & 0x3FC) + ((j >> 8) & 0x02)) = *(uint16 *)(temp + j);
+              }
+            }
+
             /* Mode 5 rendering */
-            parse_satb = parse_satb_m5;
             update_bg_pattern_cache = update_bg_pattern_cache_m5;
             if (im2_flag)
             {
+              parse_satb = parse_satb_m5_im2;
               render_bg = (reg[11] & 0x04) ? render_bg_m5_im2_vs : render_bg_m5_im2;
               render_obj = (reg[12] & 0x08) ? render_obj_m5_im2_ste : render_obj_m5_im2;
             }
             else
             {
+              parse_satb = parse_satb_m5;
               render_bg = (reg[11] & 0x04) ? (config.enhanced_vscroll ? render_bg_m5_vs_enhanced : render_bg_m5_vs) : render_bg_m5;
               render_obj = (reg[12] & 0x08) ? render_obj_m5_ste : render_obj_m5;
             }
@@ -1816,6 +1859,19 @@ static void vdp_reg_w(unsigned int r, unsigned int d, unsigned int cycles)
           }
           else
           {
+            /* 64K->16K address decoding */
+            for (i=0; i<0x4000; i+=0x400)
+            {
+              /* make temporary copy of 1KB VRAM chunk */
+              memcpy(temp, vram + i, 0x400);
+
+              /* re-arrange VRAM addressing (first and last words are unchanged) */
+              for (j=2; j<0x3fe; j+=2)
+              {
+                *(uint16 *)(vram + i + ((j >> 1) & 0x1FE) + ((j << 8) & 0x200)) = *(uint16 *)(temp + j);
+              }
+            }
+
             /* Mode 4 rendering */
             parse_satb = parse_satb_m4;
             update_bg_pattern_cache = update_bg_pattern_cache_m4;
@@ -1921,6 +1977,13 @@ static void vdp_reg_w(unsigned int r, unsigned int d, unsigned int cycles)
     {
       reg[5] = d;
       satb = (d << 9) & sat_base_mask;
+
+      /* Sprite Attribute Table Base changed during HBLANK (Madoh Monogatari I) */
+      if ((v_counter < bitmap.viewport.h) && (reg[1] & 0x40) && (cycles <= (mcycles_vdp + 860)))
+      {
+        /* re-parse sprites for next line */
+        parse_satb(v_counter);
+      }
       break;
     }
 
@@ -2377,8 +2440,8 @@ static void vdp_68k_data_w_m4(unsigned int data)
   }
   else
   {
-    /* VRAM address (interleaved format) */
-    int index = ((addr << 1) & 0x3FC) | ((addr & 0x200) >> 8) | (addr & 0x3C00);
+    /* VRAM address (16KB) */
+    int index = addr & 0x3FFE;
 
     /* Pointer to VRAM */
     uint16 *p = (uint16 *)&vram[index];
